#ifndef GRAPH_H_
#define GRAPH_H_

#include "../include/macros.h"
#include "../include/exception.h"
#include "../include/matrix.h"
#include "../include/bytebuffer.h"

#include <list>
#include <sstream>

using namespace std;

namespace simplis
{

	/*!
	 *	This class implements a graph-section in a graph that
	 *	has a small minimum and maximum vertex-degree. The 
	 *	graph-section is induced by a fixed anchor-vertex and
	 *	the radius of the section.
	 */
	class Graph
	{
		
	public:
		//! The number of vertices in the graph.
		unsigned int vertexCount;
	
		//! The index of the first node of the outermost orbit.
		unsigned int outerOrbitOffset;
	
		//! Store the number of anonymous edges for each vertex.
		unsigned int* anonymousEdgeCount;
	
		//! Store the adjacency matrix.
		Matrix* adjacencyMatrix;
	
	protected:
		//! The parameterless constructor creates the start-graph.
		Graph();
	
		//! This method allocates resources.
		void allocateMemory();
	
		//! This method releases resources.
		void releaseMemory();
	
		//! Resizes the graph.
		void resize(const unsigned int& vertexCount);	
	
	public:	
		//! Create a copy of the graph.
		Graph* copy() const;

		//! Is an edge present?
		bool edge(const unsigned int& u, const unsigned int& v);
	
		//! The destructor releases acquired resources.
		virtual ~Graph();

		//! Fetch the degree of a vertex.
		int degree(const unsigned int& i);

		//! Sets an edge.
		void setEdge(const unsigned int& i, const unsigned int& j);
	
		//! Removes an edge.
		void unsetEdge(const unsigned int& i, const unsigned int& j);
	
		//! The default constructor creates the starting graph.
		Graph(const unsigned int& vertexCount);
	
		//! Retrieve the orbit of a vertex.
		unsigned int getOrbit(const unsigned int& v);
	
		//! Sets the number of anonymous edges for a given vertex.
		void setAnonymousEdgeCount(const unsigned int& vertex, const unsigned int& count);
	
		//! Determine the radius of the graph.
		int getRadius();
		
		//! Retrieve the number of vertices in the graph.
		unsigned int size() const;
		
		//! Convert the graph into a string.
		string toString() const;

		//! Realises the anonymous edges of a graph.
		void realise();
		
	};
	
}

#endif /*GRAPH_H_*/
