#ifndef MATRIX_H_
#define MATRIX_H_

#include "../include/macros.h"
#include "../include/exception.h"

#include <string>
#include <sstream>

using namespace std;

namespace simplis
{
	
	//! A bitmatrix-class to encode graph-adjacency.
	/*!
	 *	This class implements a generic adjacency matrix. Using
	 *	bit-coding the information can be stored efficiently and
	 *	many operations can be reduced to integer-based binary
	 *	operations.
	 */
	class Matrix
	{
		
	private:
		//! How big is the matrix currently.
		unsigned int usedBits;
	
		//! How many bits are allocated?
		unsigned int allocatedBits;
	
		//! The data for the matrix.
		bool** bitmap;
		
	protected:
		//! The parameterless constructor is hidden.
		Matrix();
	
	public:
		//! The default constructor needs the matrix' size.
		Matrix(const unsigned int& size);
	
		//! The destructor releases resources.
		~Matrix();
	
		//! Print the matrix to the screen.
		string toString() const;

		//! Check a position in the matrix.
		bool get(const unsigned int& i, const unsigned int& j) const; //, qword* map = 0) const;
	
		//! Set a bit in the matrix.
		void set(const unsigned int& i, const unsigned int& j);
	
		//! Unset a bit in the matrix.
		void unset(const unsigned int& i, const unsigned int& j);
	
		//! Increase the size of the matrix.
		void enlarge(const unsigned int& size);

		//! Shrinks the matrix by one row/column.
		void shrink();

		//! Check if two vertices are foldable.
		bool foldable(const unsigned int& v, const unsigned int& w) const;
		
		//! Check if two vertices are from the same class.
		bool equivalent(const unsigned int& v, const unsigned int& w) const;
		
		//! Merge the neighborhood of w to v.
		void merge(const unsigned int& v, const unsigned int& w);
		
		//! Swaps two vertices.
		void swap(const unsigned int& v, const unsigned int& w);
		
		//! Retrieve the bitcount for a certain row.
		unsigned int degree(const unsigned int& v) const;
		
		//! Retrieve the size of the matrix.
		const unsigned int size() const;

		//! Retrieve the outer bitcount for a certain row.
		unsigned int orbit(const unsigned int& v, const unsigned int& offset);
		
		
		unsigned int count();
		
	};
	
}

#endif /*MATRIX_H_*/
