#ifndef __STACK_H
#define __STACK_H

// Use exceptions.
#include "exception.h"

namespace simplis
{
	
	template < typename KeyType >
	class Stack
	{
		
	private:
		//! Some ugly relations.
		friend class Set;
		
		//! A node in the stack.
		struct Node
		{
			//! The predecessor.
			Node* previous;
			
			//! The stored key.
			KeyType key;
		};
		
		//! The top element of the stack.
		Node* top;
		
	protected:
		
	public:
		//! The constructor creates an empty stack.
		Stack();
	
		//! The destructor clears the stack.
		~Stack();
	
		//! Push an element on the stack.
		inline void push(const KeyType& element);
	
		//! Pop an element from the stack.
		inline KeyType pop();
	
		//! Check if the stack is empty.
		inline bool empty() const;
	
		//! Peek the first element.
		inline KeyType peek() const;
		
	};
	
}

/*!
 *	The default constructor creates an empty stack.
 */
template < typename KeyType >
simplis::Stack<KeyType>::Stack()
{
	// The initial stack is empty.
	top = 0;
}

/*!
 *	This method stores an element on top of the current
 *	stack. The element should be duplicateable without
 *	side-effects.
 *
 *	@param element The new element.
 */
template < typename KeyType >
inline void simplis::Stack<KeyType>::push(const KeyType& element)
{
	// Wrap the element in a node.
	Node* newNode = new Node;
	
	// [Assertion] The allocation was successful.
	assert(newNode);
	
	// Initialize the node.
	newNode->key = element;
	newNode->previous = top;
	
	// Put the new node ontop of the stack.
	top = newNode;
}

/*!
 *	Pop an element from the stack. The element is removed
 *	from the stack by this command.
 *
 *	@return The key stored on the former top element of
 *			of the stack.
 */
template < typename KeyType >
inline KeyType simplis::Stack<KeyType>::pop()
{
	// [Assertion] The stack is not empty.
	assert(!empty());
	
	// Extract the result.
	KeyType result = top->key;
	
	// Remove the old node.
	Node* oldTop = top;
	top = oldTop->previous;
	delete oldTop;
	
	// Return the result.
	return result;
}

/*!
 *	This method checks wether the stack is empty.
 *
 *	@return <code>true</code> iff the stack is empty.
 */
template < typename KeyType >
inline bool simplis::Stack<KeyType>::empty() const
{
	return (top == 0);
}

/*!
 *	This method retrieves the top element from the stack
 *	without removing it from the stack.
 *
 *	@return The key from the top of the stack.
 */
template < typename KeyType >
inline KeyType simplis::Stack<KeyType>::peek() const
{
	// [Assertion] The stack is not empty.
	assert(top);
	
	// Return the key from the top element.
	return top->key;
}

/*!
 *	The destructor removes any remaining object from the
 *	stack.
 */
template < typename KeyType >
simplis::Stack<KeyType>::~Stack()
{
	// While the stack is not empty.
	while (!empty())
	{
		// Remove the first element.
		pop();
	}
}

#endif
