/*
 *	FINDISO-Tool
 *
 *	This tool is part of the simplis toolkit. In this tool
 *	a file of input-graphs is considered. The tool tries
 *	to find pairs of isomorphic graphs. Whenever such a
 *	pair is found, the index of one of the graphs i written
 *	into the outputfile. Thus, by removing every i-th graph,
 *	where i is contained in the output file, we can clear
 *	the inputset from any isomorphic graphs.
 */

// External dependencies.
#include <string>
#include <sstream>
#include <iostream>
#include <time.h>
#include <fstream>
#include <list>
#include "../include/graph.h"
#include "../include/graphcollection.h"
#include "../include/bytebuffer.h"

// Default namespaces.
using namespace std;
using namespace simplis;

// Global variables.
string inputFile = "";
string reportFile = "";
string outputFile = "";
time_t startTime = static_cast<time_t>(0);

// Prototypes.
bool parseArguments(int, char**);
void printSyntax();

// Point of entry.
int main(int argc, char** argv)
{
	// Try to parse the arguments.
	if (!parseArguments(argc, argv))
	{
		printSyntax();
		return 0;
	}
	
	// Print an initial message.
	startTime = time(0);

	// Open the reportfile.
	ifstream report;
	report.open(reportFile.c_str(), std::ios::out);
	
	// Open the inputfile.
	GraphCollection* dirtyFile = new GraphCollection(inputFile, GraphCollection::INPUT_COLLECTION);
	
	// Open the outputfile.
	GraphCollection* cleanFile = new GraphCollection(outputFile, GraphCollection::OUTPUT_COLLECTION);
	
	// Keep track of the dirty & clean graphs.
	unsigned long long dirtyCount = 0;
	unsigned long long cleanCount = 0;
	
	// Keep track of the next obsolete graph.
	unsigned long long nextObsolete = 0xFFFFFFFFFFFFFFFFULL;
	
	// Iterate over all dirty graphs.
	try 
	{
		while (true)
		{
			// Do we need to look up what graph is obsolete?
			if (nextObsolete == 0xFFFFFFFFFFFFFFFFULL && !report.eof()) report >> nextObsolete;

			// Retrieve the next graph.
			Graph* graph = dirtyFile->read();

			// Is this graph obsolete?
			if (nextObsolete == dirtyCount)
			{
				nextObsolete = 0xFFFFFFFFFFFFFFFFULL;
			}
			else
			{
				++cleanCount;
				cleanFile->write(*graph);
			}

			// Count the graph.
			++dirtyCount;

			// Get rid of the graph.
			delete graph;
		}
	}
	catch (int interrupt)
	{
		// Close the reportfile.
		report.close();
	
		// Remove the input file.
		delete dirtyFile;
	
		// Remove the output file.
		delete cleanFile;
	}

	return 0;	
}

void printSyntax()
{
	cerr 	<< "Invocation: sclean [OPTIONS]" << endl 
			<< "Locates pairs of isomorphic graphs and writes their index into a file." << endl
			<< endl
			<< "Arguments mandatory for long options are also mandatory for short options." << endl
			<< "  -i, --input     The graph collection to clean." << endl
			<< "  -r, --report    The report file from the isomorphy-tracker." << endl
			<< "  -o, --output    The desired output-file." << endl;
}

bool parseArguments(int argc, char** argv)
{
	// We need at least two arguments (plus the first argument that is an OS-default).
	if (argc != 4) 
	{
		return false;
	}
	
	// In case we have enough arguments, we parse every one of them.
	for (int k = 1; k < argc; ++k)
	{
		// Convert the argument into a key and a value.
		string text = argv[k];
		size_t separatorPos = text.find_first_of('=');
		string key = text.substr(1, separatorPos - 1);
		string valueText = text.substr(separatorPos + 1);
		istringstream value(valueText);
		
		// Check the key and extract the value accordingly.
		if (key == "i" || key == "-input")
		{
			value >> inputFile;
		}
		else if (key == "r" || key == "-report")
		{
			value >> reportFile;
		}
		else if (key == "o" || key == "-output")
		{
			value >> outputFile;
		}
	}

	// Any vital parameter not given?
	if (inputFile == "" || reportFile == "" || outputFile == "")
	{
		return false;
	}
	return true;
}
