/*
 *	EDGE-Tool
 *
 *	This tool is part of the simplis toolkit. In this tool
 *	the edges on the outer orbit are filled in. Therefore
 *	we can limit the intra-orbital edges by considering
 *	a minimum and maximum degree that no vertex must
 *	exceed.
 */

// External dependencies.
#include <string>
#include <sstream>
#include <iostream>
#include <time.h>
#include <fstream>
#include <list>
#include <map>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include "../include/berko.h"
#include "../include/graph.h"
#include "../include/graphcollection.h"

// Hardcoded limits.
#define MAX_ORBITAL_VERTICES 7

// Default namespaces.
using namespace std;
using namespace simplis;

// Global variables.
int maxDegree = -1;
string inputPath = "";
string outputPath = "";
unsigned long long graphsProduced = 0;
time_t startTime = static_cast<time_t>(0);
BerkowitzHash H(DEFAULT_CAP_EXP);
map<string, GraphCollection*> collections;

// Datastructures.
struct VertexPair
{
	unsigned char first;
	unsigned char second;
	static bool compare(const VertexPair& P, const VertexPair& Q)
	{
		if (P.first < Q.first)
			return true;
		if (P.first == Q.first && P.second < Q.second)
			return true;
		return false;
	}
};

// Prototypes.
bool parseArguments(int, char**);
bool is_dir(const string& path);
void printSyntax();
void edge(GraphCollection* output, const unsigned long long& counter, Graph* graph);
list<VertexPair> getPairs(Graph* graph);
void edge(GraphCollection* collection, Graph* graph, list<VertexPair> pairs);

// Point of entry.
int main(int argc, char** argv)
{
	// Try to parse the arguments.
	if (!parseArguments(argc, argv))
	{
		printSyntax();
		return 0;
	}
	
	// Print an initial message.
	startTime = time(0);

	// Try to open the source directory.
	DIR* dir = opendir(inputPath.c_str());
	if (dir == 0)
	{
		cerr << "SEDGE could not open the directory " << inputPath.c_str() << endl;
		return 0;
	}

	// Create a list of input files.
	list<string> inputFiles;

	// Get all the files from the directory.
	struct dirent* file;
	while ((file = readdir(dir)) != 0)
	{
		// Create the full path.
		string path = inputPath;
		if (inputPath[inputPath.size() - 1] != '/') path += '/';
		path += file->d_name;

		// Is the entity a directory?
		if (!is_dir(path))
		{
			inputFiles.push_back(path);
		}
	}

	// We're done with the directories.
	closedir(dir);

	// Create a graph collection to store the graphs.
	GraphCollection collection(outputPath, GraphCollection::OUTPUT_COLLECTION);

	// Iterate through all the available input files.
	for (list<string>::iterator in = inputFiles.begin(); in != inputFiles.end(); ++in)
	{
		// Determine the filename.
		string filename = *in;
		
		// Tell which file is being processed.
		cout << "SEDGE processing file \"" << filename << "\"." << endl;
		
		// Open the graph file.
		GraphCollection* input = new GraphCollection(filename, GraphCollection::INPUT_COLLECTION);

		// Keep track of the graphs extracted from that file.
		unsigned long long localGraphs = 0;

		// Try to fetch all graphs out of that file.
		try
		{
			while (true)
			{
				Graph* graph = input->read();
				
				// Prune graphs of radius=2 which have to many orbital vertices.
				if (graph->outerOrbitOffset > 1)
				{
					if (graph->vertexCount - graph->outerOrbitOffset > MAX_ORBITAL_VERTICES)
					{
						delete graph;
						continue;
					}
				}
			
				// Count the graph.
				++localGraphs;

				// Call the edge-algorithm on the input graph.
				edge(&collection, localGraphs, graph);
				
				delete graph;
			}
		}
		catch (int interrupt)
		{
			if (interrupt != PREMATURE_EOF_EXCEPTION) cerr << "ERROR: Unknown error while reading a collection." << endl;
		}

		// Release the input file.
		delete input;
	}

	// Remove the graph-writers.
	for (map<string,GraphCollection*>::iterator I = collections.begin(); I != collections.end(); ++I)
	{
		delete (*I).second;
	}
	
	// Print a result message.
	time_t duration = (time(0) - startTime);
	cout << "SEDGE guessed " << graphsProduced << " graph(s) in " << duration << "s." << endl << endl;
	return 0;	
}

void printSyntax()
{
	cerr 	<< "Invocation: sedge [OPTIONS]" << endl 
			<< "Folds the orbital vertices of a set of graphs." << endl
			<< endl
			<< "Arguments mandatory for long options are also mandatory for short options." << endl
			<< "  -i, --input     The graph collection to fold." << endl
			<< "  -o, --output    A file to store the connected graphs." << endl
			<< "  -M, --maxdegree The maximum degree of each vertex." << endl;
}

bool parseArguments(int argc, char** argv)
{
	// We need at least two arguments (plus the first argument that is an OS-default).
	if (argc != 4) 
	{
		return false;
	}
	
	// In case we have enough arguments, we parse every one of them.
	for (int k = 1; k < argc; ++k)
	{
		// Convert the argument into a key and a value.
		string text = argv[k];
		size_t separatorPos = text.find_first_of('=');
		string key = text.substr(1, separatorPos - 1);
		string valueText = text.substr(separatorPos + 1);
		istringstream value(valueText);
		
		// Check the key and extract the value accordingly.
		if (key == "i" || key == "-input")
		{
			value >> inputPath;
		}
		else if (key == "o" || key == "-output")
		{
			value >> outputPath;
		}
		else if (key == "M" || key == "--maxdegree")
		{
			value >> maxDegree;
		}
	}
	
	// Any vital parameter not given?
	if (inputPath == "" || outputPath == "" || maxDegree == -1)
	{
		return false;
	}
	return true;
}

list<VertexPair> getPairs(Graph* graph)
{
	// We start with an empty list.
	list<VertexPair> combinablePairs;
	
	// A buffer pair.
	VertexPair P;

	// Check all vertices on the outer orbit.
	for (unsigned int i = graph->outerOrbitOffset; i < graph->vertexCount; ++i)
	{
		// Is the first vertex' degree not yet satisfied?
		if (graph->adjacencyMatrix->degree(i) < static_cast<unsigned int>(maxDegree))
		{
			// Exploit symmetries.
			for (unsigned int j = i + 1; j < graph->vertexCount; ++j)
			{
				// Is the second vertex' degree not yet satisfied?
				if (graph->adjacencyMatrix->degree(i) < static_cast<unsigned int>(maxDegree))
				{
					// Store the pair (i,j) in the list of combinable pairs.
					P.first = i;
					P.second = j;
					combinablePairs.push_back(P);
				}
			}
		}
	}
	
	// Return the list in its current state.
	return combinablePairs;
}

bool is_dir(const string& path)
{
	DIR* dir = opendir(path.c_str());
	if (dir)
	{
		closedir(dir);
		return true;
	}
	else
	{
		return false;
	}
}

void edge(GraphCollection* collection, const unsigned long long& counter, Graph* graph)
{
	// Get the initial list of combineable vertices.
	list<VertexPair> pairs = getPairs(graph);
	
	// Combine all vertex-pairs.
	edge(collection, graph, pairs);
}

void edge(GraphCollection* collection, Graph* graph, list<VertexPair> pairs)
{
	// [Assertion] The graph is valid.
	assert(graph);
	
	// [Assertion] The pool is valid.
	assert(collection);

	// Case I: We ran out of edges.
	if (pairs.empty())
	{
		// Count the written graph.
		extern unsigned long long graphsProduced;
		++graphsProduced;
		
		// Hash the graph.
		string hash = H.applyHash(graph->adjacencyMatrix);
		
		// Fetch a graphcollection to write the graph to.
		GraphCollection* out;
		
		// Do we already have a writer for graphs with this hash-value?
		if (collections.find(hash) != collections.end())
		{
			// Write the graph using this writer.
			out = collections[hash];
		}
		else
		{
			// What is the output filename?
			string outFile = outputPath + '/' + hash;
			
			// Open an outputcollection to this file.
			out = new GraphCollection(outFile, GraphCollection::OUTPUT_COLLECTION);
			
			// Store the collection in the map.
			collections[hash] = out;
		}
		
		// Write the graph using the output collection
		out->write(*graph);
		return;
	}

	// Case II: We have at least one edge that we can still branch over.
	VertexPair e = *pairs.begin();
	pairs.erase(pairs.begin());
	
	// Let's assume, we would add this edge between (v,w). It is possible, that
	// v,w has the maximum allowed degree after adding that edge. Let's check on that.
	unsigned int firstDegree = graph->adjacencyMatrix->degree(e.first) + 1;
	unsigned int secondDegree = graph->adjacencyMatrix->degree(e.second) + 1;

	// Duplicate the edge list for one branch. In case the maximum degree of some vertices
	// is reached in that branch, we neglect the candidates in the list when duplicating it.
	list<VertexPair> yesCaseEdges;
	for (list<VertexPair>::iterator I = pairs.begin(); I != pairs.end(); ++I)
	{
		if ((*I).first == e.first && firstDegree == static_cast<unsigned int>(maxDegree))
		{
			continue;
		}
		if (((*I).first == e.second || (*I).second == e.second) && secondDegree == static_cast<unsigned int>(maxDegree))
		{
			continue;
		}
		yesCaseEdges.push_back(*I);
	}
	
	// Duplicate the graph.
	Graph* yesGraph = graph->copy();
	yesGraph->setEdge(e.first, e.second);
	
	// Branch into the no-case.
	edge(collection, graph, pairs);
	
	// Branch into the yes-case.
	edge(collection, yesGraph, yesCaseEdges);
	
	// Remove the duplicated graph.
	delete yesGraph;
}
