/*
 *	EXPAND-Tool
 *
 *	This tool is part of the simplis toolkit. In this tool
 *	a file of input-graphs is considered. The tool expands
 *	all graphs in the set. The minimum and maximum degree
 *	of every vertex is also considered during that process.
 */

// External dependencies.
#include <string>
#include <sstream>
#include <iostream>
#include <time.h>
#include <fstream>
#include <list>
#include <vector>
#include "../include/graph.h"
#include "../include/graphcollection.h"
#include "../include/bytebuffer.h"

// Default namespaces.
using namespace std;
using namespace simplis;

// Global variables.
string inputFile = "";
string outputFile = "";
int minDegree = -1;
int maxDegree = -1;
unsigned long long droppedByDomination = 0ULL;
unsigned long long overallGenerated = 0ULL;
time_t startTime = static_cast<time_t>(0);

// Prototypes.
bool parseArguments(int, char**);
void printSyntax();
list<Graph*> expandGraph(Graph*);
void expandGraphR(Graph*, list<Graph*>*, const unsigned int&);

// Point of entry.
int main(int argc, char** argv)
{
	// Try to parse the arguments.
	if (!parseArguments(argc, argv))
	{
		printSyntax();
		return 0;
	}
	
	// Print an initial message.
	cout << "SEXPAND expanding set..." << endl;
	startTime = time(0);
	
	// Open the graph file.
	GraphCollection* input = new GraphCollection(inputFile, GraphCollection::INPUT_COLLECTION);
	
	// Open the output file.
	GraphCollection* output = new GraphCollection(outputFile, GraphCollection::OUTPUT_COLLECTION);
	
	// Iterate the graphs in the set.
	try
	{
		while (true)
		{
			// Fetch a new graph.
			Graph* graph = input->read();

			// Expand that graph into a set of new graphs.
			list<Graph*> expanded = expandGraph(graph);

			// Write all these graphs into the output file.
			for (list<Graph*>::iterator I = expanded.begin(); I != expanded.end(); ++I) {
				output->write(**I);
				++overallGenerated;
				delete *I;
			}

			// Release the last graph.
			delete graph;
		}
	}
	catch (int interrupt)
	{
		if (interrupt != PREMATURE_EOF_EXCEPTION) cerr << "ERROR: Unknown error while reading a collection." << endl;
	}

	// Close the input set.
	delete input;
	
	// Close the output file.
	delete output;
	
	// Print a result message.
	time_t duration = (time(0) - startTime);
	cout << "SEXPAND expanded " << overallGenerated << " graph(s) in " << duration << " seconds." << endl << endl;
	return 0;	
}

void printSyntax()
{
	cerr 	<< "Invocation: sclean [OPTIONS]" << endl 
			<< "Locates pairs of isomorphic graphs and writes their index into a file." << endl
			<< endl
			<< "Arguments mandatory for long options are also mandatory for short options." << endl
			<< "  -i, --input     The graph collection to clean." << endl
			<< "  -o, --output    The desired output-file." << endl
			<< "  -m, --mindegree The minimum degree of a vertex." << endl
			<< "  -M, --maxdegree The maximum degree of a vertex." << endl;
}

bool parseArguments(int argc, char** argv)
{
	// We need at least two arguments (plus the first argument that is an OS-default).
	if (argc != 5) 
	{
		return false;
	}
	
	// In case we have enough arguments, we parse every one of them.
	for (int k = 1; k < argc; ++k)
	{
		// Convert the argument into a key and a value.
		string text = argv[k];
		size_t separatorPos = text.find_first_of('=');
		string key = text.substr(1, separatorPos - 1);
		string valueText = text.substr(separatorPos + 1);
		istringstream value(valueText);
		
		// Check the key and extract the value accordingly.
		if (key == "i" || key == "-input")
		{
			value >> inputFile;
		}
		else if (key == "o" || key == "-output")
		{
			value >> outputFile;
		}
		else if (key == "m" || key == "-mindegree")
		{
			value >> minDegree;
		}
		else if (key == "M" || key == "-maxdegree")
		{
			value >> maxDegree;
		}
	}

	// Any vital parameter not given?
	if (inputFile == "" || outputFile == "" || minDegree == -1 || maxDegree == -1)
	{
		return false;
	}
	return true;
}

list<Graph*> expandGraph(Graph* graph)
{
	// Validate the presence of the graph.
	assert(graph);
	
	// Create an initially empty list of expanded graphs.
	list<Graph*> expandedGraphs;
	
	// Use a recursive algorithm to determine the expansions of the graph.
	expandGraphR(graph->copy(), &expandedGraphs, graph->outerOrbitOffset);
	
	// Return the resulting graphs.
	return expandedGraphs;
}

void expandGraphR(Graph* graph, list<Graph*>* results, const unsigned int& next)
{
	// Validate the presence of the arguments
	assert(graph && results);
	
	// Are we done deciding on the number of anonymous edges?
	if (next >= graph->vertexCount)
	{
		// Realise the graph.
		graph->realise();

		// Store the graph in the result-pool and terminate the recursion.
		results->push_back(graph);
		return;
	}

	// Determine the degree of the next expandable vertex.
	unsigned int d = graph->adjacencyMatrix->degree(next);

	// Determine the minimum and maximum number of edges required to
	// satisfy the minimum and maximum degree criterion of the current
	// vertex.
	unsigned int min = static_cast<int>(d) > minDegree ? 0 : minDegree - d;
	unsigned int max = static_cast<int>(d) <= maxDegree ? maxDegree - d : 0;

	// For every possible number of anonymous edges, we branch into a subcase.
	for (unsigned int e = min; e <= max; ++e)
	{
		// Create a copy of the current graph.
		Graph* child = graph->copy();
		
		// Set the number of anonymous edges in the graph.
		child->anonymousEdgeCount[next] = e;
		
		// Branch into the subcase.
		expandGraphR(child, results, next + 1);
	}
	
	// Because all subcases use their own copies of our graph, we can
	// delete the input graph here.
	delete graph;
}
