/*
 *	FINDISO-Tool
 *
 *	This tool is part of the simplis toolkit. In this tool
 *	a file of input-graphs is considered. The tool tries
 *	to find pairs of isomorphic graphs. Whenever such a
 *	pair is found, the index of one of the graphs i written
 *	into the outputfile. Thus, by removing every i-th graph,
 *	where i is contained in the output file, we can clear
 *	the inputset from any isomorphic graphs.
 */

// External dependencies.
#include <string>
#include <sstream>
#include <iostream>
#include <time.h>
#include <fstream>
#include <list>
#include <vector>
#include "../include/stack.h"
#include "../include/graph.h"
#include "../include/graphcollection.h"

// Default namespaces.
using namespace std;
using namespace simplis;

// Definitions.
#ifdef MAX_DEGREE
#undef MAX_DEGREE
#endif
#define MAX_DEGREE 8
#ifdef UNDEFINED
#undef UNDEFINED
#endif
#define UNDEFINED -1

// Debugging output from isomorphy checks?
//#define DEBUG_ISO

// Global variables.
string inputFile = "";
string reportFile = "";
unsigned long long isomorphiesFound = 0;
unsigned long long overallTestCount = 0;
unsigned long long stepsPerTest = 0xffffffffffffffffULL;
unsigned long long currentSteps = 0ULL;
unsigned long long timeLimit = 0xffffffffffffffffULL;
unsigned long long startIndex = 0;
pair< list<unsigned int>, list<unsigned int> > grid[3 * MAX_DEGREE * MAX_DEGREE];
time_t startTime = static_cast<time_t>(0);
GraphCollection* firstFile = 0;
GraphCollection* secondFile = 0;
long long secondIndex = 0LL;

// Prototypes.
bool parseArguments(int, char**);
void printSyntax();
bool isomorphic(Graph*, Graph*);
bool isomorphic(Graph*, Graph*, int*);
void rewindSecondFile();
Graph* getNextGraph();
Graph* getNthGraph(const long long& n);
bool areIsomorphic(Graph* X, Graph* Y);

// Point of entry.
int main(int argc, char** argv)
{
	// Try to parse the arguments.
	if (!parseArguments(argc, argv))
	{
		printSyntax();
		return 0;
	}
	
	// Print an initial message.
	startTime = time(0);

	// Open the reportfile.
	ofstream report;
	report.open(reportFile.c_str(), std::ios::out);

	// Open the graphfile twice.
	GraphCollection* firstFile = new GraphCollection(inputFile, GraphCollection::INPUT_COLLECTION);
	secondFile = new GraphCollection(inputFile, GraphCollection::INPUT_COLLECTION);

	// Keep track of the index of the first graph.
	long long firstIndex = -1;
	
	// Check, how many graphs are stored in the file.
	long long count = 0LL;
	{
		GraphCollection* file = new GraphCollection(inputFile, GraphCollection::INPUT_COLLECTION);
		try 
		{
			while (true)
			{
				delete file->read();
				++count;
			}
		}
		catch (int interrupt)
		{
			if (interrupt != PREMATURE_EOF_EXCEPTION) cerr << "ERROR: Unknown error while reading a collection." << endl;
		}
		delete file;
		file = 0;
	}
	
	// Iterate over all graph-pairs.
	Graph* firstGraph = 0;
	try 
	{
		// Loop through the file.
		while (true)
		{
			// Extract the next graph we want to check for isomorphic duplicates.
			firstGraph = firstFile->read();
			++firstIndex;

			// Retrieve the first relevant graph.
			Graph* secondGraph = getNthGraph(firstIndex + 1);

			// Loop through the file for the second graph.
			try
			{
				// Perform isomorphy checks until we have checked all graphs or have found an isomorphic graph.
				while (true)
				{
					// Should not be necessary, but I'll make sure.
					if (firstIndex == secondIndex)
					{
						cerr << "EXCEPTION!" << endl;
						continue;
					}
				
					// Are the graphs isomorphic?
					if (isomorphic(firstGraph, secondGraph))
					{
						// Stop searching.
						break;
					}
					else
					{
						// Load the next graph from the file.
						delete secondGraph;
						secondGraph = 0;
						secondGraph = getNextGraph();
					}
				}
			}
			catch(int interrupt)
			{
			}
			
			// Have we found an isomorphic graph before?
			if (secondIndex < count)
			{
				// Store the index of the graph and count the result.
				report << firstIndex << endl;
				++isomorphiesFound;
			}

			// Remove both graphs.
			delete firstGraph;
			delete secondGraph;
			firstGraph = 0;
			secondGraph = 0;
		}
	}
	catch (int interrupt)
	{
		if (interrupt != PREMATURE_EOF_EXCEPTION) cerr << "ERROR: Unknown error while reading a collection." << endl;	
	}

	// Close the reportfile.
	report.close();
	
	// Delete the last graph.
	delete firstGraph;

	// Release the input file.
	delete firstFile;

	// Print a result message.
	return 0;	
}

bool areIsomorphic(Graph* X, Graph* Y)
{
	return false;
}

void rewindSecondFile()
{
	if (secondFile != 0) secondFile->rewind();
	secondIndex = 0ULL;
}

Graph* getNthGraph(const long long& n)
{
	// Scan from the beginning of the file.
	rewindSecondFile();
	
	// Scan until the n-th graph is found.
	Graph* last = 0;
	do
	{
		// Release the last graph.
		delete last;
		last = 0;
		
		// Get the next graph.
		last = secondFile->read();
		
		// Increase the number of the graph.
		++secondIndex;
	}
	while (secondIndex < n + 1);
	--secondIndex;
	return last;
}

Graph* getNextGraph()
{
	++secondIndex;
	return secondFile->read();
}

void printSyntax()
{
	cerr 	<< "Invocation: sfindiso [OPTIONS]" << endl 
		<< "Locates pairs of isomorphic graphs and writes their index into a file." << endl
		<< endl
		<< "Arguments mandatory for long options are also mandatory for short options." << endl
		<< "  -i, --input     The graph collection to search." << endl
		<< "  -r, --report    A file to report the located isomorphisms to." << endl
		<< "  -T, --time      A global time limit. After this time has passed, the tool stops" << endl
		<< "                  looking for isomorphic graphs." << endl
		<< "  -t, --steps	  The number of non-deterministic steps the isomorphism checking algorithm" << endl
		<< "                  is limited to for each pair of graphs." << endl
		<< "  -o, --offset    The index of the first graph to check." << endl;
}

bool parseArguments(int argc, char** argv)
{
	// We need at least two arguments (plus the first argument that is an OS-default).
	if (argc < 3) 
	{
		return false;
	}
	
	// In case we have enough arguments, we parse every one of them.
	for (int k = 1; k < argc; ++k)
	{
		// Convert the argument into a key and a value.
		string text = argv[k];
		size_t separatorPos = text.find_first_of('=');
		string key = text.substr(1, separatorPos - 1);
		string valueText = text.substr(separatorPos + 1);
		istringstream value(valueText);
		
		// Check the key and extract the value accordingly.
		if (key == "i" || key == "-input")
		{
			value >> inputFile;
		}
		else if (key == "r" || key == "-report")
		{
			value >> reportFile;
		}
		else if (key == "o" || key == "-offset")
		{
			value >> startIndex;
		}
		else if (key == "t" || key == "-steps")
		{
			value >> stepsPerTest;
		}
		else if (key == "T" || key == "-time")
		{
			value >> timeLimit;
		}
	}

	// Any vital parameter not given?
	if (inputFile == "" || reportFile == "")
	{
		return false;
	}
	return true;
}

bool isomorphic(Graph* G, Graph* H)
{	
	// Assert the presence of both graphs.
	assert(G != 0 && H != 0);
	
	// Reset the number of performed steps
	extern unsigned long long currentSteps;
	currentSteps = 0ULL;

	#ifdef DEBUG_ISO
		cout << "Isomorphy-Test for G=" << G->toString() << endl << "and H=" << H->toString() << endl;
	#endif
	
	// Do some early out.
	if (G->vertexCount != H->vertexCount || G->outerOrbitOffset != H->outerOrbitOffset) 
	{
		#ifdef DEBUG_ISO
			cout << "--> false (early-out)" << endl;
		#endif
		return false;
	}
	
	// Prepare the initial mapping (0->0, otherwise undefined)
	int* map = new int[G->vertexCount];
	map[0] = 0;
	for (unsigned int k = 1; k < G->vertexCount; ++k) map[k] = UNDEFINED;
	
	// Use the recursive algorithm to determine wether the graphs are isomorphic.
	bool result = isomorphic(G, H, map);
	
	// Remove the map from memory.
	delete [] map;
	
	// Return the result
	#ifdef DEBUG_ISO
		cout << (result ? "--> true" : "--> false") << endl;
	#endif	
	return result;
}

void printMap(Graph* G, int* map)
{
	cout << "f:G->H,v->[";
	bool first = true;
	for(unsigned int k = 0; k < G->vertexCount; ++k)
	{
		if (first) { first = false; } else { cout << ','; }
		cout << map[k];
	}
	cout << ']' << endl;
}

bool consistent(Graph* G, Graph* H, int* map, unsigned int v, unsigned int w)
{
	// For every vertex x != v, xv in E(G) <-> f(x)f(v) = f(x)w in E(H).
	for (unsigned int x = 1; x < G->vertexCount; ++x)
	{
		if (x == v) continue;	
		if (map[x] == UNDEFINED) continue;
		if (G->adjacencyMatrix->get(x,v) && !H->adjacencyMatrix->get(map[x],w)) return false;
		if (!G->adjacencyMatrix->get(x,v) && H->adjacencyMatrix->get(map[x],w)) return false;
	}
	return true;
}

bool isomorphic(Graph* G, Graph* H, int* map)
{
	// Have we done enough?
	extern unsigned long long currentSteps;
	if (currentSteps >= stepsPerTest)
	{
		return false;
	}
	else
	{
		currentSteps++;
	}

	// Are we done with the mapping?
	#ifdef DEBUG_ISO
		cout << endl;
		printMap(G, map);
	#endif	
	bool done = true;
	for (unsigned int k = 0; k < G->vertexCount; ++k) if (map[k] == UNDEFINED) done = false;
	if (done) return true;
	
	// Get a vertex that is currently unmapped.
	unsigned int v = 1;
	while (map[v] != UNDEFINED) ++v;
	#ifdef DEBUG_ISO
		cout << "Vertex " << v << " is unmapped." << endl;
	#endif
	
	// Get the degree and orbit of v in G.
	int d = G->adjacencyMatrix->degree(v);
	int o = v < G->outerOrbitOffset ? 1 : 2;
	#ifdef DEBUG_ISO
		cout << "Vertex " << v << " is of degree " << d << " and orbit " << o << "." << endl;
	#endif
	
	// For all vertices w in H of the same orbit & degree as v in G, try to map v->w.
	for (unsigned int w = 1; w < H->vertexCount; ++w)
	{
		// Check if w is not used in the current mapping.
		bool unmapped = true;
		for (unsigned int k = 1; k < G->vertexCount; ++k) if (map[k] == static_cast<int>(w)) unmapped = false;
		
		// If w is already used we cannot map v->w, because we need the map to be injective.
		if (!unmapped) continue;

		// Check wether w has the same degree and orbit as v.
		int dw = H->adjacencyMatrix->degree(w);
		int ow = w < H->outerOrbitOffset ? 1 : 2;
		
		// If v,w do not match in orbit and degree, they may not be mapped onto each other.
		if (d != dw || o != ow) continue;
		
		#ifdef DEBUG_ISO
			cout << "We could try " << v << "->" << w << " [d(v)=" << d << ",d(w)=" << dw << ",o(v)=" << o << ",o(w)=" << ow << "]." << endl;
		#endif
		
		// Check wether mapping v->w would keep the map consistent.
		if (consistent(G, H, map, v, w))
		{
			// map+v->w may yield a possible solution. Create the new map and check it recursively.
			#ifdef DEBUG_ISO
				cout << "Mapping " << v << "->" << w << " is consistent." << endl;
			#endif
			int* new_map = new int[G->vertexCount];
			for (unsigned int k = 0; k < G->vertexCount; ++k) new_map[k] = map[k];
			new_map[v] = w;
			
			// Check if this map yields a solution.
			bool solution = isomorphic(G, H, new_map);
			
			// Delete the new map.
			delete [] new_map;
			
			// Have we found a solution?
			if (solution) return true;
		}		
		#ifdef DEBUG_ISO
			else
			{
				cout << "Mapping " << v << "->" << w << " will compromise the mappings consistency!" << endl;
			}
		#endif
	}
	return false;
}
























