/*
 *	FOLD-Tool
 *
 *	This tool is part of the simplis toolkit. The fold-tool loads
 *	a graphcollection from disk and parses it graph by graph. For
 *	Every graph the possible foldings of the orbital veritces is
 *	determined using 
 */

// External dependencies.
#include <string>
#include <sstream>
#include <iostream>
#include <time.h>
#include <stdio.h>
#include <fstream>
#include <list>
#include "../include/graph.h"
#include "../include/graphcollection.h"
#include "../include/bytebuffer.h"

// Default namespaces.
using namespace std;
using namespace simplis;

// Global variables.
int minDegree = -1;
int maxDegree = -1;
string inputPath = "";
string outputPath = "";
unsigned long long graphsProduced = 0;
time_t startTime = static_cast<time_t>(0);

// Datastructures.
struct VertexPair
{
	unsigned char first;
	unsigned char second;
	static bool compare(const VertexPair& P, const VertexPair& Q)
	{
		if (P.first < Q.first)
			return true;
		if (P.first == Q.first && P.second < Q.second)
			return true;
		return false;
	}
};

// Prototypes.
bool parseArguments(int, char**);
void printSyntax();
void printCandidates(list<VertexPair>&);
void fold(const string& output, const unsigned long long& counter, Graph* graph);
void fold(GraphCollection* collection, Graph* graph, list<VertexPair> pairs);
list<VertexPair> getPairs(Graph* graph);

// Point of entry.
int main(int argc, char** argv)
{
	// Try to parse the arguments.
	if (!parseArguments(argc, argv))
	{
		printSyntax();
		return 0;
	}
	
	// Print an initial message.
	cout << "SFOLD folding graphset..." << endl;
	startTime = time(0);
		
	// Open the graph file.
	GraphCollection* input = new GraphCollection(inputPath, GraphCollection::INPUT_COLLECTION);
	
	// Parse the whole file.
	try
	{
		while (true)
		{
			Graph* graph = input->read();
			fold(outputPath, graphsProduced, graph); // fold the current graph.
			delete graph;
		}
	}
	catch (int interrupt)
	{
		if (interrupt != PREMATURE_EOF_EXCEPTION) cerr << "ERROR: Unknown error while reading a collection." << endl;
	}
	
	// Cleanup.
	delete input;
	
	// Print a result message.
	time_t duration = (time(0) - startTime);
	cout << "SFOLD folded " << graphsProduced << " graph(s) in " << duration << " seconds." << endl << endl;
	return 0;	
}

void printSyntax()
{
	cerr 	<< "Invocation: sfold -i=input -o=output" << endl 
			<< "Folds the orbital vertices of a set of graphs." << endl
			<< endl
			<< "Arguments mandatory for long options are also mandatory for short options." << endl
			<< "  -i, --input     The graph collection to fold." << endl
			<< "  -o, --output    A file to store the folded graphs." << endl;
}

bool parseArguments(int argc, char** argv)
{
	// We need at least two arguments (plus the first argument that is an OS-default).
	if (argc != 3) 
	{
		return false;
	}
	
	// In case we have enough arguments, we parse every one of them.
	for (int k = 1; k < argc; ++k)
	{
		// Convert the argument into a key and a value.
		string text = argv[k];
		size_t separatorPos = text.find_first_of('=');
		string key = text.substr(1, separatorPos - 1);
		string valueText = text.substr(separatorPos + 1);
		istringstream value(valueText);
		
		// Check the key and extract the value accordingly.
		if (key == "i" || key == "-input")
		{
			value >> inputPath;
		}
		else if (key == "o" || key == "-output")
		{
			value >> outputPath;
		}
	}
	
	// Any vital paramter not given?
	if (inputPath == "" || outputPath == "")
	{
		return false;
	}
	return true;
}

void fold(const string& output, const unsigned long long& count, Graph* graph)
{
	// Is the graph valid?
	assert(graph != 0);
	
	// At first we locate all possibly foldable vertexpairs from the initial graph.
	list<VertexPair> pairs = getPairs(graph);
	
	// Create the output file for the tree of the current graph.
	stringstream s;
	s << output;
	if (output[output.size() - 1] != '/')
	{
		s << '/';
	}
	s << count;

	// Create a graph collection to store the graphs.
	GraphCollection collection(s.str(), GraphCollection::OUTPUT_COLLECTION);

	// Generate all foldings.
	fold(&collection, graph, pairs);
}

void printCandidates(list<VertexPair>& pairs) {
	bool first = true;
	for (list<VertexPair>::iterator i = pairs.begin(); i != pairs.end(); ++i) {
		if (first) {
			first = false;
		} else {
			cout << ",";
		}
		cout << "(" << static_cast<int>((*i).first) << "-" << static_cast<int>((*i).second) << ")";
	}
	cout << endl;
}

list<VertexPair> getPairs(Graph* graph)
{
	// We start with an empty list.
	list<VertexPair> foldablePairs;
	
	// A buffer pair.
	VertexPair P;
	
	// Check all vertices on the outer orbit.
	for (unsigned int i = graph->outerOrbitOffset; i < graph->vertexCount; ++i)
	{
		// Exploit symmetries.
		for (unsigned int j = i + 1; j < graph->vertexCount; ++j)
		{
			// Are the orbital vertices i,j foldable?
			if (graph->adjacencyMatrix->foldable(i,j))
			{
				// Store the pair (i,j) in the list of foldable pairs.
				P.first = i;
				P.second = j;
				foldablePairs.push_back(P);
			}
		}
	}

//	printCandidates(foldablePairs);
	
	// Return the list in its current state.
	return foldablePairs;
}

//! Evaluate all folding strategies.
void fold(GraphCollection* collection, Graph* graph, list<VertexPair> pairs)
{	
//	printCandidates(pairs);

	// In case we have no further foldable pairs, we insert the current
	// into the pool and terminate.
	if (pairs.empty())
	{
		// Write the graph.
		collection->write(*graph);
		++graphsProduced;
		return;
	}
	
	// Take the first pair (x,y) from the list of foldable pairs.
	VertexPair pair = *pairs.begin();
	
	/*
	 *	If we do not fold (x,y) we will not fold any (v,w) where v,w are
	 *	from the same classes as x,y.
	 */
	list<VertexPair> notFoldedPairs;
	for (list<VertexPair>::const_iterator I = pairs.begin(); I != pairs.end(); ++I)
	{
		// In case a vertex pair is not from the same class, we note it
		// as foldable in the following branch.
		if (!graph->adjacencyMatrix->equivalent((*I).first, pair.first) 
			|| !graph->adjacencyMatrix->equivalent((*I).second, pair.second))		
		{
			notFoldedPairs.push_back(*I);
		}
	}

	/*
	 *	Consider the case that we fold (x,y). In this case we have to
	 *	do two things: Manipulate the graph and modify the pairlist.
	 */
	
	// Create a new graph to fold.
	Graph* F = graph->copy();
	
	// By construction x is lexicographically smaller than y. Hence
	// we fold y into x.
	F->adjacencyMatrix->merge(pair.first, pair.second);
	
	// By construction our vertex y is obsolete now. It is much easier
	// to delete y if y is the vertex with the highest id. Thus we
	// may need to swap that maximum vertex with y.
	if (static_cast<unsigned int>(pair.second) + 1 != F->adjacencyMatrix->size())
	{
		// Swap the vertices in the graph while preserving connectivity
		// information.
		F->adjacencyMatrix->swap(pair.second, F->adjacencyMatrix->size() - 1);
		
		// We need to swap the vertex labels in the pairs list.
		unsigned int last = F->adjacencyMatrix->size() - 1;
		for (list<VertexPair>::iterator I = pairs.begin(); I != pairs.end(); ++I)
		{
			// Need to substitute the first vertex?
			if ((*I).first == pair.second)
			{
				(*I).first = last;
			} 
			else if ((*I).first == last)
			{
				(*I).first = pair.second;
			}
			
			// What about the second?
			if ((*I).second == pair.second)
			{
				(*I).second = last;
			}
			else if ((*I).second == last)
			{
				(*I).second = pair.second;
			}
			
			// After a substitution the intra-pair order may be
			// messed up.
			if ((*I).first > (*I).second)
			{
				unsigned int temp = (*I).first;
				(*I).first = (*I).second;
				(*I).second = temp;
			}
		}
		
		// Our vertex y has been relabeled. We need to be aware of that.
		pair.second = last;
	}
	
	// Now the obsolete vertex is the maximum vertex. We remove it by
	// simply shrinking the matrix' size by 1.
	F->adjacencyMatrix->shrink();
	F->vertexCount -= 1;
	
	// At this point the former vertices x,y do not exist anymore.
	// Thus we can remove all pairs that involve these vertices
	// somehow (We cannot merge things we do not have).
	list<VertexPair>::iterator I = pairs.begin();
	while (I != pairs.end())
	{
		if ((*I).first == pair.first || (*I).first == pair.second || (*I).second == pair.first || (*I).second == pair.second)
		{
			// Remove the element and forward the iterator.
			I = pairs.erase(I);
		}
		else
		{
			// Manually forward the iterator.
			++I;
		}
	}
	
	// Our recently merged vertex may be foldable again. Up next is to
	// decide with which vertices it can be merged.
	VertexPair introduce;
	introduce.second = pair.first;
	for (unsigned int i = F->outerOrbitOffset; i < pair.first; ++i)
	{
		if (F->adjacencyMatrix->foldable(i, pair.first))
		{
			introduce.first = i;
			pairs.push_back(introduce);
		}
	}
	introduce.first = pair.first;
	for (unsigned int i = pair.first + 1; i < F->vertexCount; ++i)
	{
		if (F->adjacencyMatrix->foldable(pair.first, i))
		{
			introduce.second = i;
			pairs.push_back(introduce);
		}
	}
	
	// The last thing we need to do, is to sort the list of pairs, so
	// that we automatically have the representative pair for two
	// classes at hand (the idea is, that a class is represented by
	// the smallest vertexlabel in the class).
	pairs.sort(VertexPair::compare);
	
	/* 
	 *	Now we are done constructing both cases. We will work on both
	 *	of them in a recursive manner.
	 */
	
	// Case I - We did not merge (x,y).
	fold(collection, graph, notFoldedPairs);

	// Case II - We merged (x,y).
	fold(collection, F, pairs);
}

