#include "../include/graph.h"

bool simplis::Graph::edge(const unsigned int& u, const unsigned int& v) {
	return adjacencyMatrix->get(u,v);
}

int simplis::Graph::degree(const unsigned int& i) {
	return adjacencyMatrix->degree(i) + anonymousEdgeCount[i];
}

/*!
 *	This method realises the anonymous edges, by
 *	adding real vertices at the ends of those edges.
 */
void simplis::Graph::realise()
{
	// By how many vertices will the graph's size increase?
	unsigned int additionalVertices = 0;
	for (unsigned int i = outerOrbitOffset; i < vertexCount; ++i)
	{
		additionalVertices += anonymousEdgeCount[i];
	}

	// Create a new adjacency matrix.
	Matrix* matrix = new Matrix(vertexCount + additionalVertices);

	// Copy the old matrix.
	for (unsigned int v = 0; v < vertexCount; ++v)
	{
		for (unsigned int w = v + 1; w < vertexCount; ++w)
		{
			if (adjacencyMatrix->get(v, w))
			{
				matrix->set(v, w);
			}
		}
	}

	// Create edges to the new vertices.
	unsigned int nextFreeVertex = vertexCount;
	unsigned int outerVertex = outerOrbitOffset;
	while (outerVertex < vertexCount)
	{
		for (unsigned int k = 0; k < anonymousEdgeCount[outerVertex]; ++k)
		{
			matrix->set(outerVertex, nextFreeVertex++);
		}
		++outerVertex;
	}

	// Substitute the current matrix.
	delete adjacencyMatrix;
	adjacencyMatrix = matrix;

	// Adjust the outer orbit offset.
	outerOrbitOffset = vertexCount;

	// Increase the vertex count.
	vertexCount += additionalVertices;

	// Create a new array for the anonymous edges.
	delete [] anonymousEdgeCount;
	anonymousEdgeCount = new unsigned int [vertexCount];
	for (unsigned int i = 0; i < vertexCount; ++i)
	{
		anonymousEdgeCount[i] = 0;
	}
}

/*!
 *	The default constructor creates the starting-graph.
 *	This is a graph with a single node and a specific amount
 *	of anonymous edges.
 *
 *	@param vertexCount the number of anonymous edges incident
 *		   			   to the anchor node.
 */
simplis::Graph::Graph(const unsigned int& vertexCount)
{
	// The initial graph consists of a single vertex.
	this->vertexCount = vertexCount;

	// Initialize the anonymous edge count as a null-reference.
	anonymousEdgeCount = 0;
	
	// Allocate the memory.
	allocateMemory();
	
	// The outermost orbit consists of the anchor vertex.
	outerOrbitOffset = 0;
	
	// The anchor orbit has the supplied number of anonymous edges.
	anonymousEdgeCount[0] = 0;
}

/*!
 *	This method adds an edge to the graph.
 */
void simplis::Graph::setEdge(const unsigned int& i, const unsigned int& j)
{
	// [Assertion] The indices are valid.
	assert(i < vertexCount && j < vertexCount);
	
	// Add the edge.
	adjacencyMatrix->set(i,j);
}

/*!
 *	This method removes an edge from the graph.
 */
void simplis::Graph::unsetEdge(const unsigned int& i, const unsigned int& j)
{
	// [Assertion] The indices are valid.
	assert(i < vertexCount && j < vertexCount);
	
	// Add the edge.
	adjacencyMatrix->unset(i,j);
	adjacencyMatrix->unset(j,i);
}

void simplis::Graph::setAnonymousEdgeCount(const unsigned int& vertex, const unsigned int& count)
{
	// [Assertion] Vertex index is in range.
	assert(vertex < vertexCount);
	
	// [Assertion] Buffer exists.
	assert(anonymousEdgeCount);
	
	// Set amount of anonymous edges.
	anonymousEdgeCount[vertex] = count;
}

/*!
 *	The parameterless constructor is hidden, so that the user
 *	may not create an empty graph.
 */
simplis::Graph::Graph()
{
}

/*!
 *	The destructor releases resources acquired by the graph.
 */
simplis::Graph::~Graph()
{
	// Release resources.
	releaseMemory();
}

/*!
 *	This method releases the memory occupied by the graph object.
 */
void simplis::Graph::releaseMemory()
{
	// Remove the counter for the anonymous edges.
	if (anonymousEdgeCount)
	{
		delete [] anonymousEdgeCount;
		anonymousEdgeCount = 0;
	}
	
	// Remove the adjacency matrix.
	if (adjacencyMatrix)
	{
		delete adjacencyMatrix;
		adjacencyMatrix = 0;
	}
}

/*!
 *	This method allocates memory after the vertex count is
 *	set.
 */
void simplis::Graph::allocateMemory()
{
	// [Assertion] We have at least one vertex in the graph.
	assert(vertexCount > 0);

	// Remove any remaining memory.
	delete [] anonymousEdgeCount;	

	// The initial vertex has 'anchorDegree' anonymous edges.
	anonymousEdgeCount = new unsigned int [vertexCount];
	for (unsigned int k = 0; k < vertexCount; ++k)
	{
		anonymousEdgeCount[k] = 0;
	}
	
	// Initialize the adjacency matrix.
	adjacencyMatrix = new Matrix(vertexCount);
}

/*!
 *	This method reallocates memory to store more vertices in the graph.
 *
 *	@param vertexCount The number of vertices in the new graph.
 */
void simplis::Graph::resize(const unsigned int& vertexCount)
{
	// [Assertion] We need at least one vertex.
	assert(vertexCount > 0);
	
	// Anything to do?
	if (vertexCount == this->vertexCount)
	{
		return;
	}
	
	// Does the size of the graph decrease?
	if (vertexCount < this->vertexCount)
	{
		// Store the new vertex count.
		this->vertexCount = vertexCount;
		return;
	}
	
	// So the size increases...
	this->vertexCount = vertexCount;

	// Increase the size of the adjacency matrix.
	adjacencyMatrix->enlarge(vertexCount);

	// Increase the number of anonymous edges.
	delete [] anonymousEdgeCount;
	anonymousEdgeCount = new unsigned int [vertexCount];
	memset(anonymousEdgeCount, 0x00, sizeof(unsigned int) * vertexCount);
}

/*!
 *	This method creates a copy of the graph under consideration.
 *
 *	@return A pointer to an exact duplicate of the graph.
 */
namespace simplis
{
	Graph* Graph::copy() const
	{
		// Create duplicate.
		simplis::Graph* copy = new Graph();

		// Initialize the null-reference in the duplicated graph.
		copy->anonymousEdgeCount = 0;
		
		// It has the same amount of vertices.
		copy->vertexCount = vertexCount;
		
		// Allocate memory for the duplicate.
		copy->allocateMemory();
		
		// Set the offset of the outermost orbit.
		copy->outerOrbitOffset = outerOrbitOffset;
		
		// Copy the anonmyous edge counter.
		for (unsigned int i = 0; i < vertexCount; ++i)
		{
			copy->anonymousEdgeCount[i] = anonymousEdgeCount[i];
		}
		
		// Copy the adjacency matrix.
		for (unsigned int i = 0; i < vertexCount; ++i)
		{
			for (unsigned int j = 0; j < vertexCount; ++j)
			{
				if (adjacencyMatrix->get(i,j)) 
				{
					copy->adjacencyMatrix->set(i,j);
				}
				else
				{
					copy->adjacencyMatrix->unset(i,j);
				}
			}
		}
		return copy;
	}
}

/*!
 *	Retrieve the size of the graph.
 *
 *	@return The number of vertices in the graph.
 */
unsigned int simplis::Graph::size() const
{
	// [Assertion] The matrix is valid.
	assert(adjacencyMatrix);
	
	// Return the size.
	return adjacencyMatrix->size();
}

string simplis::Graph::toString() const
{
	stringstream s;
	s << "GRAPH={" << endl << "\tV={0,...," << vertexCount - 1 << "}," << endl << "\tE={";
	bool first = true;
	for (unsigned int v = 0; v < vertexCount; ++v)
	{
		for (unsigned int w = v + 1; w < vertexCount; ++w)
		{
			if (adjacencyMatrix->get(v,w))
			{
				if (first) { first = false; } else { s << ','; }
				s << '(' << v << '-' << w << ')';
			}
		}
	}
	s << '}' << endl << "\tOE=[";
	first = true;
	for (unsigned int v = 0; v < vertexCount; ++v)
	{
		if (anonymousEdgeCount[v] > 0)
		{
			if (first) { first = false; } else { s << ','; }
			s << v << "=>" << anonymousEdgeCount[v];
		}
	}
	s << "]," << endl << "\tOFF=" << outerOrbitOffset << endl << "}";
	return s.str();
}

unsigned int simplis::Graph::getOrbit(const unsigned int& v)
{
	assert(v < vertexCount);
	if (v == 0) return 0;
	if (outerOrbitOffset == 1) return 1;
	return (v >= outerOrbitOffset ? 2 : 1);
}
