#include "../include/graphcollection.h"

simplis::GraphCollection::GraphCollection(string file, simplis::GraphCollection::CollectionMode mode)
{
	// Which mode do we use?
	switch (mode)
	{
	case INPUT_COLLECTION:
		this->mode = GraphCollection::INPUT_COLLECTION;
		break;
	case OUTPUT_COLLECTION:
		this->mode = GraphCollection::OUTPUT_COLLECTION;
		break;
	default:
		throw exception("Unknown GraphCollection::ColletionMode!");
	}
	
	// Allocate the necessary bit-manipulator.
	if (mode == INPUT_COLLECTION)
	{
		reader = new ByteReader(const_cast<char*>(file.c_str()));
	}
	else
	{
		outName = file;
	}
}

simplis::Graph* simplis::GraphCollection::read(int o2align)
{
	// Verify read privileges.
	if (mode != INPUT_COLLECTION)
	{
		throw exception("Cannot read from a graph collection of type OUTPUT_COLLECTION!");
	}
	#ifdef READ_DEBUG
		cout << "Reading graph from file..." << endl;
	#endif
	
	// Create the adjacency matrix of the graph. Limited by the collection's
	// binary encoding, there can only be 100 vertices at max.
	Matrix matrix(100);
	
	// Use an array to store the anonymous edges.
	unsigned int anonymousEdges[100];
	for (int i = 0; i < 100; ++i) anonymousEdges[i] = 0;
	
	// Retrieve the radius of the graph.
	unsigned int radius = 1 + reader->readNumeric(1);
	#ifdef READ_DEBUG
		cout << "Radius=" << radius << endl;
	#endif
	
	// Read the number of vertices on the first orbit.
	unsigned int o1vertices = 1 + reader->readNumeric(3);
	#ifdef READ_DEBUG
		cout << "Vertices on first orbit=" << o1vertices << endl;
	#endif
	
	// Hence we know the number of inner vertices and wether
	// the graph has radius 2 or not, we already can determine
	// the outer orbit offset.
	unsigned int outerOrbitOffset = (radius == 1) ? 1 : 1 + o1vertices;
	#ifdef READ_DEBUG
		cout << "OuterOrbitOffset=" << outerOrbitOffset << endl;
	#endif
	
	// Keep track of the greatest vertex known.
	unsigned int maxIndex = o1vertices;
	
	// Iterate through the vertices on the first orbit.
	for (unsigned int v = 1; v <= o1vertices; ++v)
	{
		// The vertex v is connected to the anchor vertex.
		matrix.set(0, v);
		
		// Read the number of edges that this vertex has inside the
		// first orbit.
		unsigned int o1edges = reader->readNumeric(3);
		#ifdef READ_DEBUG
			cout << "o1edges for " << v << "=" << o1edges << endl << flush;
		#endif
		
		// Extract all those edges.
		for (unsigned int e = 0; e < o1edges; ++e)
		{
			// To which vertex does this edge connect?
			unsigned int other = reader->readNumeric(3) + 1;
			
			// Add the edge to the matrix.
			matrix.set(v, other);
			
			// Keep track of the maximum index.
			if (other > maxIndex) maxIndex = other;
		}
		
		// In case the radius is 1, we need to check how many
		// anonymous edges are incident to this vertex.
		if (radius == 1)
		{
			anonymousEdges[v] = reader->readNumeric(3);
		}
		else
		{
			// To how many vertices in the second orbit is this vertex
			// adjacent?
			unsigned int o2neighbors = reader->readNumeric(3);
			
			// Get the neighbors on the second orbit.
			for (unsigned int w = 0; w < o2neighbors; ++w)
			{
				// Get the index of the neighbor.
				unsigned int o2neighbor = outerOrbitOffset + reader->readNumeric(o2align);
				
				// Connect them with an edge.
				matrix.set(v, o2neighbor);
				
				// Keep track of the greatest vertex-index seen so far.
				if (o2neighbor > maxIndex) maxIndex = o2neighbor;				
			}
		}
	}

	// In case the radius of the graph is 2, we will load the edges
	// that connect vertices both from the second orbit.
	if (radius == 2)
	{
		for (unsigned int v = outerOrbitOffset; v <= maxIndex; ++v)
		{
			// Get the number of edges incident to this vertex.
			unsigned int o2edges = reader->readNumeric(3);
			
			// Extract all the edges.
			for (unsigned int e = 0; e < o2edges; ++e)
			{
				// Load the edge from the file.
				unsigned int o2target = reader->readNumeric(o2align) + outerOrbitOffset;
				
				// Store the edge in the matrix.
				matrix.set(v, o2target);
			}
			
			// Retrieve the number of anonymous edges.
			unsigned int outerEdges = reader->readNumeric(3);
			anonymousEdges[v] = outerEdges;
		}
	}
	
	// Is the graph invalid?
	if (maxIndex == 0 || radius == 0)
	{
		return 0;
	}
	#ifdef READ_DEBUG
		cout << "Maximum vertex=" << maxIndex << endl;
	#endif
	
	// Create a graph of the required size.
	Graph* G = new Graph(maxIndex + 1);	
	
	// Set the outer orbit offset.
	G->outerOrbitOffset = outerOrbitOffset;
	
	// Set the anonymous edges.
	for (unsigned int v = outerOrbitOffset; v <= maxIndex; ++v)
	{
		G->anonymousEdgeCount[v] = anonymousEdges[v];
	}

	// Set the edges.
	for (unsigned int v = 0; v <= maxIndex; ++v)
	{
		for (unsigned int w = v + 1; w <= maxIndex; ++w)
		{
			bool vw = matrix.get(v,w);
			if (vw == true)
			{
				G->setEdge(v,w);
			}
		}
	}
	
	// Skip the graph's padding in the file
	reader->skipPadding();
	
	// Return the extracted graph.	
	return G;
} 

void simplis::GraphCollection::write(const Graph& graph, int o2align)
{
	// Verify write privileges.
	if (mode != OUTPUT_COLLECTION)
	{
		throw exception("Cannot write to a graph collection of type INPUT_COLLECTION!");
	}

	// Create a byte buffer to write the binary coding.
	ByteBuffer coding;
	
	// Write the radius of the graph.
	unsigned int radius = (graph.outerOrbitOffset > 1 ? 2 : 1);
	coding.writeNumeric(radius - 1, 1);

	// Write the number of vertices on the first orbit.
	int orbitSize = (radius == 1) ? graph.vertexCount - 1: graph.outerOrbitOffset - 1;
	coding.writeNumeric(orbitSize - 1, 3);
		
	// Iterate through all vertices on orbit one.
	unsigned int limit = 1 + orbitSize;
	for (unsigned int v = 1; v < limit; ++v)
	{
		// Write the number of orbital edges incident to that vertex.
		unsigned int intraOrbitCount = 0;
		for (unsigned int w = v + 1; w < limit; ++w)
		{
			if (v != w && graph.adjacencyMatrix->get(v,w))
			{
				++intraOrbitCount;
			}
		}
		coding.writeNumeric(intraOrbitCount, 3);
			
		// Write the index of every orbit 1 vertex that is adjacent to v.
		for (unsigned int w = v + 1; w < limit; ++w)
		{
			if (v != w && graph.adjacencyMatrix->get(v,w))
			{
				coding.writeNumeric(w - 1, 3);
			}
		}
		
		// Has the graph no second orbit?
		if (radius == 1)
		{
			// We just store the number of anonymous edges for this vertex.
			coding.writeNumeric(graph.anonymousEdgeCount[v], 3);
		}
		
		// Is there a second orbit?
		if (radius == 2)
		{
			// Write the number of edges to the second orbit.
			unsigned int outerEdgeCount = 0;
			for (unsigned int w = graph.outerOrbitOffset; w < graph.vertexCount; ++w)
			{
				if (graph.adjacencyMatrix->get(v, w))
				{
					++outerEdgeCount;
				}
			}
			coding.writeNumeric(outerEdgeCount, 3);
			
			// Write the edges to the writer.
			for (unsigned int w = graph.outerOrbitOffset; w < graph.vertexCount; ++w)
			{
				if (graph.adjacencyMatrix->get(v, w))
				{
					coding.writeNumeric(w - graph.outerOrbitOffset, o2align);
				}
			}
		}
	}
		
	// Do we have an outer orbit?
	if (radius == 2)
	{
		// Iterate through the vertices on the second orbit.
		for (unsigned int v = graph.outerOrbitOffset; v < graph.vertexCount; ++v)
		{
			// Write the number of intra orbital edges incident to v.
			unsigned int intraEdges = 0;
			for (unsigned int w = graph.outerOrbitOffset; w < graph.vertexCount; ++w)
			{
				if (v != w && graph.adjacencyMatrix->get(v, w))
				{
					++intraEdges;
				}
			}
			coding.writeNumeric(intraEdges, 3);
			
			// Store the edges.
			for (unsigned int w = graph.outerOrbitOffset; w < graph.vertexCount; ++w)
			{
				if (v != w && graph.adjacencyMatrix->get(v, w))
				{
					coding.writeNumeric(w - graph.outerOrbitOffset, o2align);
				}
			}
			
			// Store the number of anonymous edges.
			coding.writeNumeric(graph.anonymousEdgeCount[v], 3);
		}
	}

	// Write the graphs to the output stream.
	coding.appendToFile(outName);
}

simplis::GraphCollection::~GraphCollection()
{
	if (mode == GraphCollection::INPUT_COLLECTION)
	{
		delete reader;
	}
}


