/*
 *	HASH-Tool
 *
 *	Takes a graphset as input. For each graph in that file
 *	the hash value is determined and the graph is appended
 *	to a file named after the hash value in a certain
 *	output directory.
 */

// External dependencies.
#include <string>
#include <sstream>
#include <iostream>
#include <time.h>
#include <fstream>
#include <list>
#include <map>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include "../include/graph.h"
#include "../include/graphcollection.h"
#include "../include/berko.h"

// Default namespaces.
using namespace std;
using namespace simplis;

// Global variables.
string inputFile = "";
string outputPath = "";
time_t startTime = static_cast<time_t>(0);

// Prototypes.
bool parseArguments(int, char**);
bool is_dir(const string& path);
void printSyntax();

// Point of entry.
int main(int argc, char** argv)
{
	// Try to parse the arguments.
	if (!parseArguments(argc, argv))
	{
		printSyntax();
		return 0;
	}

	// Open the source file.
	GraphCollection collection(inputFile, GraphCollection::INPUT_COLLECTION);

	// Create a hasher.
	BerkowitzHash* hasher = new BerkowitzHash(DEFAULT_CAP_EXP);

	// Iterate through all the available graphs.
	unsigned long long count = 0;
	try
	{
		while (true)
		{
			// Determine the name of the next file.
			stringstream S;
			S << outputPath << '/';
			
			// Read the next graph.
			Graph* next = collection.read();

			// Increase the numbering.
			++count;

			// Determine the hash value of the graph.
			string hash = hasher->applyHash(next->adjacencyMatrix);

			// Append the graph to a new collection.
			S << hash;

			// Output the graph to a new graph collection.
			GraphCollection out(S.str(), GraphCollection::OUTPUT_COLLECTION);
			out.write(*next);

			// Remove the graph from memory.
			delete next;
		}
	}	
	catch (int interrupt)
	{
	}	

	// Release the hasher.
	delete hasher;
	hasher = 0;	

	// Print a result message.
	return 0;	
}

void printSyntax()
{
	cerr 	<< "Invocation: shash [OPTIONS]" << endl 
		<< "Splits a set according to the hash values of each graph." << endl
		<< endl
		<< "Arguments mandatory for long options are also mandatory for short options." << endl
		<< "  -i, --input     The input file." << endl
		<< "  -o, --output    A file to store the hash files." << endl;
}

bool parseArguments(int argc, char** argv)
{
	// We need at least two arguments (plus the first argument that is an OS-default).
	if (argc != 3) 
	{
		return false;
	}
	
	// In case we have enough arguments, we parse every one of them.
	for (int k = 1; k < argc; ++k)
	{
		// Convert the argument into a key and a value.
		string text = argv[k];
		size_t separatorPos = text.find_first_of('=');
		string key = text.substr(1, separatorPos - 1);
		string valueText = text.substr(separatorPos + 1);
		istringstream value(valueText);
		
		// Check the key and extract the value accordingly.
		if (key == "i" || key == "-input")
		{
			value >> inputFile;
		}
		else if (key == "o" || key == "-output")
		{
			value >> outputPath;
		}
	}
	
	// Any vital parameter not given?
	if (inputFile == "" || outputPath == "")
	{
		return false;
	}
	return true;
}

bool is_dir(const string& path)
{
	DIR* dir = opendir(path.c_str());
	if (dir)
	{
		closedir(dir);
		return true;
	}
	else
	{
		return false;
	}
}
