/*
 *	INIT-Tool
 *
 *	This tool is part of the simplis toolkit. The init-tool
 *	creates the inital graphs that are the source of the
 *	graph generation process. See the built-in syntax
 *	explanation for information on the tool's usage.
 */

// External dependencies.
#include <string>
#include <sstream>
#include <iostream>
#include <time.h>
#include "../include/graph.h"
#include "../include/graphcollection.h"

// Default namespaces.
using namespace std;
using namespace simplis;

// Global variables.
int minDegree = -1;
int maxDegree = -1;
string outputFile = "";
int graphsProduced = 0;
time_t startTime = static_cast<time_t>(0);

// Prototypes.
bool parseArguments(int, char**);
void printSyntax();

// Point of entry.
int main(int argc, char** argv)
{
	// Try to parse the arguments.
	if (!parseArguments(argc, argv))
	{
		printSyntax();
		return 0;
	}
	
	// Try to create the output file.
	GraphCollection* set = new GraphCollection(outputFile, GraphCollection::OUTPUT_COLLECTION);

	// Generate the graphs.
	for (int d = minDegree; d <= maxDegree; ++d)
	{
		// Create a d-star graph.
		Graph* G = new Graph(d+1);
		for (int e = 1; e <= d; ++e)
		{
			G->setEdge(0, e);
			G->setAnonymousEdgeCount(e, 0);
			G->outerOrbitOffset = 1;
		}
		
		// Store the graph using the collection.
		set->write(*G);
		++graphsProduced;
		
		// Release the graph from memory.
		delete G;
	}

	// Cleanup.
	delete set;
	return 0;	
}

/*!
 *	Print the command line explanation of the the invocation syntax.
 */
void printSyntax()
{
	cerr 	<< "Invocation: sinit [OPTIONS]" << endl 
		<< "Generates a set of initial graphs for the generation process." << endl
		<< endl
		<< "Arguments mandatory for long options are also mandatory for short options." << endl
		<< "  -m, --mindegree    Specifies the minimum degree of the anchorvertex." << endl
		<< "  -M, --maxdegree    Specifies the maximum degree of the anchorvertex." << endl
		<< "  -o, --output       Selects the output file (defaults to 'init.out')" << endl << endl;
}

/*!
 *	Parse the supplied command line arguments and extract the required information.
 *	In case some mandatory parameters are not present or malformed the function
 *	indicates this by its return value.
 *
 *	@param argc The number of arguments passed to the program.
 *	@param argv A vector of C-strings that contain the command line arguments.
 *	@return <code>true</code> whether all mandatory arguments are given and correctly formatted.
 */
bool parseArguments(int argc, char** argv)
{
	// We need at least two arguments (plus the first argument that is an OS-default).
	if (argc < 3) 
	{
		return false;
	}
	
	// In case we have enough arguments, we parse every one of them.
	for (int k = 1; k < argc; ++k)
	{
		// Convert the argument into a key and a value.
		string text = argv[k];
		size_t separatorPos = text.find_first_of('=');
		string key = text.substr(1, separatorPos - 1);
		string valueText = text.substr(separatorPos + 1);
		istringstream value(valueText);
		
		// Check the key and extract the value accordingly.
		if (key == "m" || key == "mindegree")
		{
			value >> minDegree;
		}
		else if (key == "M" || key == "maxdegree")
		{
			value >> maxDegree;
		}
		else if (key == "o" || key == "output")
		{
			value >> outputFile;
		}
	}
	
	// Any vital paramter not given?
	if (minDegree == -1 || maxDegree == -1)
	{
		return false;
	}
	return true;
}
