#include "../include/matrix.h"

/*!
 *	The default constructor is hidden so that no instance
 *	that is only partially specified may be created.
 */
simplis::Matrix::Matrix()
{
}

/*!
 *	The default constructor creates a matrix with a given
 *	size.
 *
 *	@param size The size of the matrix.
 */
simplis::Matrix::Matrix(const unsigned int& size)
{
	// [Assertion] The size is at least 1 (empty graphs are not supported)!
	assert(size > 0);
	
	// Store the size.
	usedBits = size;
	
	// Create a fresh bitmap only if the size of the matrix is greater than 1.
	if (size == 1)
	{
		// We do not allocate a bitmap at all.
		bitmap = 0;
		
		// We did not allocate anything.
		allocatedBits = 0;
	}
	else
	{
		// Allocate the memory.
		bitmap = new bool*[size];
		for (unsigned int k = 0; k < size; ++k) 
		{
			bitmap[k] = new bool[size];
			for (unsigned int l = 0; l < size; ++l) bitmap[k][l] = false;
		}
		
		// [Assertion] Allocation was successful.
		assert(bitmap);
	}
}

/*!
 *	The destructor releases the allocated memory.
 */
simplis::Matrix::~Matrix()
{
	// Release the bitmatrix.
	if (bitmap)
	{
		for (unsigned int k = 0; k < usedBits; ++k)
		{
			delete [] bitmap[k];
			bitmap[k] = 0;
		}
		delete [] bitmap;
		bitmap = 0;
	}
}

/*!
 *	This method checks, wether a bit in the matrix is set.
 *
 *	@param i The row of the matrix.
 *	@param j The column of the matrix.
 *	@return <code>true</code> iff the bit (i,j) is set.
 */
#include <iostream>
using namespace std;
bool simplis::Matrix::get(const unsigned int& i, const unsigned int& j) const
{
	// [Assertion] Parameters are in a valid range.
	assert(i < usedBits && j < usedBits);
	
	// The diagonal of the matrix consist only of zeroes.
	if (i == j)
	{
		return false;
	}
	
	assert(bitmap[i][j] == bitmap[j][i]);

	//cout << "Matrix::get(" << i << "," << j << ") with usedBits=" << usedBits << endl << flush;
	return bitmap[i][j];
}

/*!
 *	This method activates a bit at a certain position.
 *
 *	@param i The row of the bit.
 *	@param j The column of the bit.
 */
void simplis::Matrix::set(const unsigned int& i, const unsigned int& j)
{
	// [Assertion] Parameters are in a valid range.
	assert(i < usedBits && j < usedBits);
	
	// The diagonal of the matrix consist only of zeroes.
	if (i == j)
	{
		return;
	}
	
	//cout << "Matrix::set(" << i << "," << j << ") with usedBits=" << usedBits << endl << flush;
	bitmap[j][i] = true;
	bitmap[i][j] = true;
}

/*!
 *	This method deactivates a bit at a certain position.
 *
 *	@param i The row of the bit.
 *	@param j The column of the bit.
 */
void simplis::Matrix::unset(const unsigned int& i, const unsigned int& j)
{
	// [Assertion] Parameters are in a valid range.
	assert(i < usedBits && j < usedBits);
	
	// The diagonal of the matrix consist only of zeroes.
	if (i == j)
	{
		return;
	}
	
	//cout << "Matrix::unset(" << i << "," << j << ") with usedBits=" << usedBits << endl << flush;
	bitmap[j][i] = false;
	bitmap[i][j] = false;
}

/*!
 *	This method prints the matrix to the screen.
 */
string simplis::Matrix::toString() const
{
	stringstream S;
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		for (unsigned int j = 0; j < usedBits; ++j)
		{
			if (get(i,j))
			{
				S << '1';
			}
			else
			{
				S << '0';
			}
		}
		S << endl;
	}
	return S.str();
}

/*!
 *	This method increases the size of the matrix.
 *
 *	@param size The new size of the matrix.
 */
void simplis::Matrix::enlarge(const unsigned int& size)
{
	// [Assertion] The size must not decrease.
	assert(size > usedBits);
	
	// Create bigger matrix.
	bool** large = new bool*[size];
	for (unsigned int k = 0; k < size; ++k)
	{
		large[k] = new bool[size];
	}

	// Initialise the new matrix.
	for (unsigned int i = 0; i < size; ++i)
	{
		for (unsigned int j = 0; j < size; ++j)
		{
			large[i][j] = (i < usedBits && j < usedBits) ? bitmap[i][j] : false;
		}
	}
	
	// Release the old map.
	for (unsigned int k = 0; k < usedBits; ++k)
	{
		delete [] bitmap[k];
		bitmap[k] = 0;
	}
	delete [] bitmap;
	
	// Use the new one.
	bitmap = large;
	
	// Remember the new size.
	usedBits = size;
}

/*!
 *	This method shrinks the matrix by discarding the vertex with the highest
 *	id.
 */
void simplis::Matrix::shrink()
{
	// [Assertion] The resulting graph will not be empty.
	assert(usedBits > 1);

	// Remember the new size.
	usedBits -= 1;
}

/*!
 *	This method checks if two vertices v,w are foldable in
 *	the sense of the term 'foldable' as it occurs in the
 *	expansion algorithm.
 *
 *	@param v The first vertex.
 *	@param w The other vertex.
 *	@return <code>true</code> iff the neighborhoods of v,w are
 *			disjunct.
 */
bool simplis::Matrix::foldable(const unsigned int& v, const unsigned int& w) const
{
	// [Assertion] The vertices v,w are in range.
	assert(v < usedBits && w < usedBits);

	// The vertices v,w are foldable, iff. the entries in their rows
	// are disjunctive.
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		if (get(i, v) && get(i, w))
		{
			return false;
		}
	}
	
	// We've found no inconsistencies.
	return true;
}

/*!
 *	This method checks wether two vertices v,w are from the same class.
 *	meaning in the context of vertex-merging, that both vertices have
 *	matching neighborhoods.
 *
 *	@param v The first vertex.
 *	@param w The second vertex.
 *	@return <code>true</code> iff the vertices are from the same class.
 */
bool simplis::Matrix::equivalent(const unsigned int& v, const unsigned int& w) const
{
	// [Assertion] Vertices are in range.
	assert(v < usedBits && w < usedBits);
	
	// [Assertion] v,w may not be connected.
	assert(!get(v,w));
	
	// Check if the neighborhoods of v,w match. By assertion v,w are
	// not connected.
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		if (get(i, v) != get(i, w))
		{
			// One of the vertices v,w is connected to i and the other is not.
			// Thus N(v) != N(w) holds.
			return false;
		}
	}
	
	// The neighborhoods are equal.
	return true;
}

/*!
 *	This method merges the neighborhood of w to the neighborhood of v.
 *	Thus every vertex that is adjacent to v or w will be adjacent
 *	to v afterwards. Formally: N[v] <- N[v] + N[w]
 *
 *	@param v The target vertex.
 *	@param w The soruce vertex.
 */
void simplis::Matrix::merge(const unsigned int& v, const unsigned int& w)
{
	// [Assertion] The merged vertices may not be connected.
	assert(!get(v,w));

	// Every neighbor of w becomes a neighbor of v.
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		if (get(i, w))
		{
			set(i, v);
		}
	}
}

/*!
 *	This method swaps two vertices. In the internal representation of the
 *	matrix.
 *
 *	@param v The first vertex.
 *	@param w The second vertex.
 */
void simplis::Matrix::swap(const unsigned int& v, const unsigned int& w)
{
	// [Assertion] The vertices are in range.
	assert(v < usedBits && w < usedBits);

	// Store the adjacency of v.
	bool* adjacency = new bool[usedBits];
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		adjacency[i] = get(i, v);
	}
	
	// Copy the adjacency of w over v.
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		if (get(i, w))
		{
			set(i, v);
		}
		else
		{
			unset(i, v);
		}
	}
	
	// Set the adjacency of w according to our records.
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		if (adjacency[i])
		{
			set(i, w);
		}
		else
		{
			unset(i, w);
		}
	}
	
	// Release the buffer.
	delete [] adjacency;
}

/*!
 *	This method returns the size of the matrix.
 *
 *	@return The size of the matrix.
 */
const unsigned int simplis::Matrix::size() const
{
	return usedBits;
}

/*!
 *	This method counts the number of bits set in the matrix. Active
 *	bits located in padded memory are not counted.
 *
 *	@return The number of active bits in the matrix.
 */
unsigned int simplis::Matrix::degree(const unsigned int& v) const
{
	// Store the results in an accumulator.
	unsigned int accumulator = 0;

	// Count active bits.
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		if (get(i, v))
		{
			++accumulator;
		}
	}
	return accumulator;	
}

/*!
 *	This method counts the neighbors a desired vertex has, that lie
 *	on the outer orbit in the graph.
 *
 *	@param row The desired row of the matrix.
 *	@param offset The offset of the outer orbit.
 *	@return The number of active bits in the row, greater than the
 *			orbital offset.
 */
unsigned int simplis::Matrix::orbit(const unsigned int& v, const unsigned int& offset)
{
	// [Assertion] The row index is in range.
	assert(v < usedBits);
	
	// Store the results in an accumulator.
	unsigned int accumulator = 0;

	// Count active bits in the desired row.
	for (unsigned int j = offset; j < usedBits; ++j)
	{
		if (get(j, v))
		{
			++accumulator;
		}
	}
	return accumulator;		
}

unsigned int simplis::Matrix::count()
{
	unsigned int c = 0;
	for (unsigned int i = 0; i < usedBits; ++i)
	{
		for (unsigned int j = 0; j < usedBits; ++j)
		{
			if (get(i,j))
			{
				++c;
			}
		}
	}
	return c;
}
